// hook.cpp - hook imported function
//
// Old Mail Arrangement Plug-in for Becky! 2
// Copyright (C) 2000 by Tietew
//
#include "pch.h"
#include "hook.h"

// Get import descriptor in a module.
static PIMAGE_IMPORT_DESCRIPTOR
	GetNamedImportDescriptor(HMODULE hModule, LPCSTR szImportMod)
{
	if(!szImportMod || !hModule)
		return NULL;
	
	// HMODULE is a pointer for IMAGE_DOS_HEADER
	PIMAGE_DOS_HEADER pDOSHeader = (PIMAGE_DOS_HEADER)hModule;
	
	if(IsBadReadPtr(pDOSHeader, sizeof IMAGE_DOS_HEADER)
	|| pDOSHeader->e_magic != IMAGE_DOS_SIGNATURE)
		return NULL;
	
	// Get IMAGE_NT_HEADERS
	PIMAGE_NT_HEADERS pNTHeader =
		(PIMAGE_NT_HEADERS)((LPBYTE)pDOSHeader + pDOSHeader->e_lfanew);
	
	if(IsBadReadPtr(pNTHeader, sizeof IMAGE_NT_HEADERS)
	|| pNTHeader->Signature != IMAGE_NT_SIGNATURE)
		return NULL;
	
	// Get array of import descriptors in the module
	PIMAGE_IMPORT_DESCRIPTOR pImportDesc =
		(PIMAGE_IMPORT_DESCRIPTOR)((LPBYTE)pDOSHeader
			+ pNTHeader->OptionalHeader.DataDirectory
				[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress);
	if(!pImportDesc)
		return NULL;
	
	// Search
	for(; pImportDesc->Name; ++pImportDesc)
		if(lstrcmpi((LPSTR)((LPBYTE)pDOSHeader + pImportDesc->Name),
				szImportMod) == 0)
			break;
	
	return pImportDesc->Name ? pImportDesc : NULL;
}

// Hook imported functions in a module
BOOL HookImportedFunctionsByName(HMODULE hModule, LPCSTR szImportMod,
	UINT uiCount, LPHOOKFUNCDESC paHookArray, PROC *paOrigFuncs,
	LPUINT puiHooked)
{
	UINT i;
	
	if(!uiCount || !szImportMod
	|| IsBadWritePtr(puiHooked, sizeof UINT)
	|| IsBadWritePtr(paOrigFuncs, sizeof PROC * uiCount)
	|| IsBadReadPtr(paHookArray, sizeof HOOKFUNCDESC * uiCount))
		return FALSE;
	
	memset(paOrigFuncs, 0, sizeof PROC * uiCount);
	*puiHooked = 0 ;
	
	// Get the import descriptor
	PIMAGE_IMPORT_DESCRIPTOR pImportDesc =
					 GetNamedImportDescriptor(hModule, szImportMod);
	if(!pImportDesc)
		return FALSE;
	
	// Get thunk data
	PIMAGE_THUNK_DATA pOrigThunk =
		(PIMAGE_THUNK_DATA)((LPBYTE)hModule + pImportDesc->OriginalFirstThunk);
	PIMAGE_THUNK_DATA pRealThunk =
		(PIMAGE_THUNK_DATA)((LPBYTE)hModule	+ pImportDesc->FirstThunk);
	
	// Search each function and hook
	while(pOrigThunk->u1.Function)
	{
		if((pOrigThunk->u1.Ordinal & IMAGE_ORDINAL_FLAG) != IMAGE_ORDINAL_FLAG)
		{
			PIMAGE_IMPORT_BY_NAME pByName = 
				(PIMAGE_IMPORT_BY_NAME)((LPBYTE)hModule
					+ (DWORD)pOrigThunk->u1.AddressOfData);
			
			if(!pByName->Name[0])
				continue;
			
			bool bDoHook = false;
			for(i = 0; i < uiCount; i++)
				if(paHookArray[i].szFunc[0] == pByName->Name[0]
				&& lstrcmpi(paHookArray[i].szFunc, (char*)pByName->Name) == 0)
				{
					if(paHookArray[i].pProc)
						bDoHook = true;
					break;
				}
			
			if(bDoHook)
			{
				MEMORY_BASIC_INFORMATION mbi_thunk;
				DWORD dwOldProtect;
				
				// Allow writing to descriptor
				VirtualQuery(pRealThunk, &mbi_thunk, sizeof MEMORY_BASIC_INFORMATION);
				VirtualProtect(mbi_thunk.BaseAddress, mbi_thunk.RegionSize,
					PAGE_READWRITE, &mbi_thunk.Protect);
				
				// Hook the function
				paOrigFuncs[i] = (PROC)pRealThunk->u1.Function;
				(DWORD&)pRealThunk->u1.Function = (DWORD)paHookArray[i].pProc;
				++*puiHooked;
				
				// Restore protection
				VirtualProtect(mbi_thunk.BaseAddress, mbi_thunk.RegionSize,
					mbi_thunk.Protect, &dwOldProtect);
			}
		}
		pOrigThunk++;
		pRealThunk++;
	}
	
	return TRUE;
}
