// remove.cpp - arrange folder
//
// Old Mail Arrangement Plug-in for Becky! 2
// Copyright (C) 2000 by Tietew
//
#include "pch.h"
#include "bkrmold.h"
#include "remove.h"
#include "utys.h"
#include "progress.h"
#include "foldercombo.h"
#include "hook.h"
#include "filter.h"

// MessageBox API that force return IDYES
static int WINAPI MessageBoxForceYes(HWND, LPCSTR, LPCSTR, UINT)
{
	return IDYES;
}

// Dialog procedure of confirmation dialog
BOOL CALLBACK ConfirmProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch(msg)
	{
		case WM_INITDIALOG:
		{
			RMINFO& info = *(RMINFO*)lp;
			SetWindowLong(hwnd, DWL_USER, lp);
			
			// Initialize controls
			int days = GetIniDays();
			SetDlgItemText(hwnd, IDC_FOLDER, bka.GetFolderDisplayName(info.name));
			SetDlgItemInt(hwnd, IDC_DAYS, days, FALSE);
			SendDlgItemMessage(hwnd, IDC_DAYSSPIN, UDM_SETRANGE32, 0, INT_MAX);
			SetDlgItemInt(hwnd, IDC_MAILS, GetIniMails(), FALSE);
			SendDlgItemMessage(hwnd, IDC_MAILSSPIN, UDM_SETRANGE32, 0, INT_MAX);
		//	CheckDlgButton2(hwnd, IDC_RECURSIVE, GetIniRecursive());
			CheckDlgButton2(hwnd, IDFromBehavior(GetIniBehavior()), TRUE);
			CheckDlgButton2(hwnd, GetIniFilter() ? IDC_FLAST : IDC_FDAYS, TRUE);
			CheckDlgButton2(hwnd, IDC_THREAD, GetIniThreads());
			
			SYSTEMTIME st[2];
			unsigned __int64 ft;
			GetLocalTime(&st[1]);
			st[0].wYear = 1970;
			st[0].wMonth = 1;
			st[0].wDay = 1;
			SendDlgItemMessage(hwnd, IDC_DATE, DTM_SETRANGE, GDTR_MIN | GDTR_MAX, (LPARAM)st);
			SystemTimeToFileTime(&st[1], (LPFILETIME)&ft);
			ft -= FT_DAY * days;
			FileTimeToSystemTime((LPFILETIME)&ft, &st[0]);
			SendDlgItemMessage(hwnd, IDC_DATE, DTM_SETSYSTEMTIME, GDT_VALID, (LPARAM)&st[0]);
			
			String fini = bka.GetDataFolder() + info.name + "Folder.ini";
			EnableWindow(GetDlgItem(hwnd, IDC_THREAD), GetPrivateProfileInt("Status", "Thread", 0, fini));
			
			// Initialize folder combobox
			char dest[MAX_PATH];
			GetPrivateProfileString(INI_SECTION, "LastDestination", "", dest, countof(dest), fini);
			if(!*dest)
				GetPrivateProfileString(INI_SECTION, "Destination", "", dest, countof(dest), fini);
			if(!InitFolderCombo(hwnd, IDC_MOVETO, *dest ? dest : bka.GetCurrentFolder()))
				EndDialog(hwnd, IDCANCEL);
			MarkMoreButton(hwnd, IDC_MORE, *info.pfi);
			
			// Initial control enableness
			PostMessage(hwnd, WM_COMMAND, IDC_SELECT, 0);
			return TRUE;
		}
		case WM_COMMAND:
			switch(LOWORD(wp))
			{
				case IDOK:
				{
					RMINFO& info = *(RMINFO*)GetWindowLong(hwnd, DWL_USER);
					if(IsDlgButtonChecked2(hwnd, IDC_FLAST))
					{
						info.nFilter = IDC_FLAST;
						info.u.nMails = GetDlgItemInt(hwnd, IDC_MAILS, NULL, FALSE);
					}
					else if(IsDlgButtonChecked2(hwnd, IDC_FDAYS))
					{
						info.nFilter = IDC_FDAYS;
						info.u.nDays = GetDlgItemInt(hwnd, IDC_DAYS, NULL, FALSE);
					}
					else
					{
						info.nFilter = IDC_FDATE;
						SendDlgItemMessage(hwnd, IDC_DATE, DTM_GETSYSTEMTIME, 0, (LPARAM)&info.u.st);
					}
					info.bRecursive = IsDlgButtonChecked2(hwnd, IDC_RECURSIVE);
					info.bThreads = IsDlgButtonChecked2(hwnd, IDC_THREAD);
					info.nBehavior = IsDlgButtonChecked2(hwnd, IDC_MOVE) ? IDC_MOVE :
						IsDlgButtonChecked2(hwnd, IDC_REMOVE) ? IDC_REMOVE : IDC_SELECT;
					if(info.nBehavior == IDC_MOVE)
					{
						info.target = GetSelectedFolder(hwnd, IDC_MOVETO);
						if(info.target == info.name || info.target == String::Nil)
						{
							// Alert and continue
							MyMessageBox(hwnd, info.target == String::Nil ? IDS_MOVETO_MAILBOX : IDS_MOVETO_ME,
								MB_OK | MB_ICONEXCLAMATION);
							SetFocus(GetDlgItem(hwnd, IDC_MOVETO));
							return TRUE;
						}
						// Remember destination
						WritePrivateProfileString(INI_SECTION, "LastDestination", info.target,
							String(bka.GetDataFolder() + info.name + "Folder.ini"));
					}
				}
				// fall through
				case IDCANCEL:
					EndDialog(hwnd, wp);
					return TRUE;
				
				case IDC_SELECT:
				case IDC_MOVE:
				case IDC_REMOVE:
				{
					// Enableness of controls
					bool bSelect = IsDlgButtonChecked2(hwnd, IDC_SELECT);
					bool bMove = IsDlgButtonChecked2(hwnd, IDC_MOVE);
					bool bRemove = IsDlgButtonChecked2(hwnd, IDC_REMOVE);
					EnableWindow(GetDlgItem(hwnd, IDC_STATIC_MOVETO), bMove);
					EnableWindow(GetDlgItem(hwnd, IDC_MOVETO), bMove);
					EnableWindow(GetDlgItem(hwnd, IDC_RECURSIVE), bRemove && false);
					return TRUE;
				}
				case IDC_DAYS:
					if(HIWORD(wp) == EN_SETFOCUS)
						CheckRadioButton(hwnd, IDC_FDAYS, IDC_FLAST, IDC_FDAYS);
					return TRUE;
				
				case IDC_MAILS:
					if(HIWORD(wp) == EN_SETFOCUS)
						CheckRadioButton(hwnd, IDC_FDAYS, IDC_FLAST, IDC_FLAST);
					return TRUE;
				
				case IDC_MORE:
				{
					RMINFO& info = *(RMINFO*)GetWindowLong(hwnd, DWL_USER);
					FILTERDLGINFO fdi;
					fdi.fi = *info.pfi;
					fdi.nCurSel = 0;
					int result = DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_FILTER),
						hwnd, (DLGPROC)FilterProc, (LPARAM)&fdi);
					if(result == IDOK)
						*info.pfi = fdi.fi;
					MarkMoreButton(hwnd, IDC_MORE, *info.pfi);
					return TRUE;
				}
			}
			break;
		
		case WM_NOTIFY:
			if(wp == IDC_DATE)
			{
				NMHDR& nmhdr = *(LPNMHDR)lp;
				if(nmhdr.code == NM_SETFOCUS || nmhdr.code == DTN_DROPDOWN)
					CheckRadioButton(hwnd, IDC_FDAYS, IDC_FLAST, IDC_FDATE);
				return TRUE;
			}
			break;
		
		case WM_MEASUREITEM:
		case WM_DRAWITEM:
			return ProcessFolderCombo(hwnd, msg, wp, lp);
		
		case WM_DESTROY:
			TermFolderCombo(hwnd, IDC_MOVETO);
			break;
	}
	return FALSE;
}

// Calculate threshold time_t value from RMINFO
DWORD CalcThresholdDate(RMINFO& ri)
{
	SYSTEMTIME st;
	unsigned __int64 ft, ft2;
	
	if(ri.nFilter == IDC_FDAYS)
	{
		GetLocalTime(&st);
		st.wHour = st.wMinute = st.wSecond = st.wMilliseconds = 0;
		SystemTimeToFileTime(&st, (LPFILETIME)&ft);
		ft -= FT_DAY * ri.u.nDays - FT_DAY;
	}
	else
	{
		st = ri.u.st;
		st.wHour = st.wMinute = st.wSecond = st.wMilliseconds = 0;
		SystemTimeToFileTime(&st, (LPFILETIME)&ft);
		ft += FT_DAY;
	}
	// Apply time zone
	LocalFileTimeToFileTime((LPFILETIME)&ft, (LPFILETIME)&ft2);
	return (DWORD)((ft2 - 116444736000000000) / FT_SECOND);
}

// Thread-Tree
typedef map<String, String> TRefMap;
// functional object that returns if left mail is newer than right.
struct date_greater : public binary_function<INDEXINFO, INDEXINFO, bool>
{
	bool operator()(const INDEXINFO& lhs, const INDEXINFO& rhs) const
		{ return lhs.dwDate > rhs.dwDate; }
};
// functional object that returns if the mail is NOT AdvancedFilter target
class not_filtered : public unary_function<INDEXINFO, bool>
{
	const FILTERINFO& fi;
public:
	not_filtered(const FILTERINFO& fi) : fi(fi) { }
	bool operator()(const INDEXINFO& idx) const
		{ return !AdvancedFilter(idx, fi); }
};
// functional object that returns if the mail is NOT thread root.
class thread_child : public unary_function<INDEXINFO, bool>
{
	const TRefMap& imap;
public:
	thread_child(const TRefMap& imap) : imap(imap) { }
	bool operator()(const INDEXINFO& idx) const
		{ return imap.find(idx.refid) != imap.end(); }
};
// functional object that returns if the mail is erased.
class erased_mail : public unary_function<INDEXINFO, bool>
{
	const TRefMap& imap;
public:
	erased_mail(const TRefMap& imap) : imap(imap) { }
	bool operator()(const INDEXINFO& idx) const
		{ return imap.find(idx.id) == imap.end(); }
};
enum { CANCELED, NOMAIL, DOARRANGE };

// erase the mail and all parent mails
template<class T> static void erase_refs(T& imap, String id)
{
	T::iterator it = imap.find(id);
	if(it != imap.end())
	{
		erase_refs(imap, it->second);
		imap.erase(it);
	}
}

int SelectMails(RMINFO& info, bool bUI)
{
	// Progress dialog
	CBeckyWindow wnd;
	CProgress prg(bUI);
	String idbase = info.name + '?';
	int i = 0;
	String line;
	INDEXINFO idxinfo;
	CFolderIndex idx(info.name);
	TRefMap imap;
	vector<INDEXINFO> v;
	vector<INDEXINFO>::iterator first, last, it;
	
	// Error Check
	if(idx.Failed() == 1)
	{
		MyMessageBox(wnd.hwndMain, IDS_FAIL_IDX, MB_ICONHAND | MB_OK);
		return CANCELED;
	}
	else if(idx.Failed() == 2)
	{
		MyMessageBox(wnd.hwndMain, IDS_BAD_IDX, MB_ICONEXCLAMATION | MB_OK);
		return CANCELED;
	}
	
	// AdvancedFilter
	if(info.pfi && info.pfi->v.size() == 0)
		info.pfi = NULL;
	
	// Scan Folder.ini and decide mails to be selected
	if(info.nFilter != IDC_FLAST)
	{
		INDEXINFO idxThreshold;
		idxThreshold.dwDate = CalcThresholdDate(info);
		
		while((line = idx.GetLine()) != String::Nil)
		{
			CFolderIndex::Parse(line, idxinfo);
			// push list and make tree
			v.push_back(idxinfo);
			imap.insert(TRefMap::value_type(idxinfo.id, idxinfo.refid));
			if((++i & 0x1F) == 0 && prg.DispatchMessage(0, idx.GetPercent()))
				return CANCELED;
		}
		if(prg.DispatchMessage(0, 100))
			return CANCELED;
		last = v.end();
		// Apply filter
		first = partition(v.begin(), last, bind2nd(date_greater(), idxThreshold));
		if(info.pfi)
			first = partition(first, last, not_filtered(*info.pfi));
		if(info.bThreads)
		{
			// remove mails that is not thread root
			last = remove_if(first, last, thread_child(imap));
			// erase unfiltered mail and its parent mails of all
			for(it = v.begin(); it != first; ++it)
				erase_refs(imap, it->refid);
			// remove erased mails
			last = remove_if(first, last, erased_mail(imap));
		}
	}
	else
	{
		while((line = idx.GetLine()) != String::Nil)
		{
			INDEXINFO idxinfo;
			CFolderIndex::Parse(line, idxinfo);
			// push list and make tree
			v.push_back(idxinfo);
			imap.insert(TRefMap::value_type(idxinfo.id, idxinfo.refid));
			if((++i & 0x1F) == 0 && prg.DispatchMessage(0, idx.GetPercent()))
				return CANCELED;
		}
		if(prg.DispatchMessage(0, 100))
			return CANCELED;
		if(v.size() <= info.u.nMails)
			return NOMAIL;
		
		// Sort order by date descend
		sort(v.begin(), v.end(), date_greater());
		first = v.begin() + info.u.nMails;
		last = v.end();
		// Apply Advanced Filter
		if(info.pfi)
			first = partition(first, last, not_filtered(*info.pfi));
		if(info.bThreads)
		{
			// remove mails that is not thread root
			last = remove_if(first, last, thread_child(imap));
			// erase unfiltered mail and its parent mails of all
			for(it = v.begin(); it != first; ++it)
				erase_refs(imap, it->refid);
			// remove erased mails
			last = remove_if(first, last, erased_mail(imap));
		}
	}
	// No mails will be selected
	if(first == last)
		return NOMAIL;
	
	// Select mails
	// Set current folder and unselect all mails
	bka.SetCurrentMail(String(idbase + first->id));
	SendMessage(wnd.hwndMain, WM_COMMAND,
		info.bThreads ? BKID_VIEW_CLOSEALLTHREADS : BKID_VIEW_OPENALLTHREADS, 0);
	prg.DisableUI();
	int max = distance(first, last), n = 0;
	for(it = first; it != last; ++it)
	{
		bka.SetSel(String(idbase + it->id), true);
		if((++n & 0x1F) == 0 && prg.DispatchMessage(1, n * 100 / max))
			return CANCELED;
	}
	if(prg.DispatchMessage(1, 100))
		return CANCELED;
	return DOARRANGE;
}

// Arrange mails
bool RemoveMails(RMINFO& info, bool bUI)
{
	CBeckyWindow wnd;
	
	// Mails to be selected
	int stat = SelectMails(info, bUI);
	if(stat != DOARRANGE)
		return stat == CANCELED;
	
	// Force activate ListView
	// ** MFC calls SetActiveView in OnMouseActivate handler
	SendMessage(wnd.hwndList, WM_MOUSEACTIVATE, (WPARAM)wnd.hwndMain,
		MAKELPARAM(HTBORDER, WM_MBUTTONDOWN));
	
	if(info.nBehavior == IDC_REMOVE)
	{
		HOOKFUNCDESC hfd[] = {
			{ "MessageBoxA", (PROC)MessageBoxForceYes },
		};
		PROC orig[countof(hfd)];
		UINT ui;
		// Hook MessageBoxA API
		HookImportedFunctionsByName(bka.GetB2Instance(), "user32.dll",
			countof(hfd), hfd, orig, &ui);
		// Shift-key down
		SHORT shift = GetAsyncKeyState(VK_SHIFT);
		keybd_event(VK_SHIFT, 0, 0, 0);
		// Delete mails without any UI
		SendMessage(wnd.hwndMain, WM_COMMAND, BKID_MAIL_DELETE, 0);
		// Restore shift-key status
		if(shift >= 0)
			keybd_event(VK_SHIFT, 0, KEYEVENTF_KEYUP, 0);
		// Unhook MessageBoxA API
		hfd[0].pProc = orig[0];
		HookImportedFunctionsByName(bka.GetB2Instance(), "user32.dll",
			countof(hfd), hfd, orig, &ui);
	}
	else if(info.nBehavior == IDC_MOVE)
		// Simply move mails
		bka.MoveSelectedMessages(info.target, false);
	return true;
}

// Menu handler of arrangement mails
void CALLBACK OnRemoveMails(HWND hWnd, LPARAM lParam)
{
	RMINFO info;
	FILTERINFO fi;
	info.name = bka.GetCurrentFolder();
	info.pfi = &fi;
	if(DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_CONFIRM), hWnd, (DLGPROC)ConfirmProc, (LPARAM)&info) != IDOK)
		return;
	if(!RemoveMails(info, true))
		MyMessageBox(hWnd, IDS_NOMAIL, MB_OK | MB_ICONINFORMATION);
}

// Update handler of arrangement mails
UINT CALLBACK OnUpdateRemoveMails(HWND hWnd, LPARAM lParam)
{
	// Disables if current folder is mailbox root
	String box = bka.GetCurrentMailBox();
	return lstrcmpi(bka.GetCurrentFolder(), box) == 0 ? BKMENU_CMDUI_DISABLED : 0;
}
